{-# LANGUAGE UndecidableInstances #-}
module Types.Value
  ( Value(..)
  , Valuable(..)
  ) where

import Data.Dynamic

data Value = Value { valueDynamic :: Dynamic
                   , valueShow :: String
                   }

instance Show Value where
  show = valueShow

class Typeable a => Valuable a where
  toValue :: a -> Value
  fromValue :: Value -> a

toValueOnce :: (Typeable a, Show a) => a -> Value
toValueOnce x = Value { valueDynamic = toDyn x
                      , valueShow = show x
                      }

fromValueOnce :: Typeable a => Value -> a
fromValueOnce = flip fromDyn (error "unexpected") . valueDynamic

instance Valuable Value where
  toValue = id
  fromValue = id

instance {-# OVERLAPPABLE #-} Valuable String where
  toValue = toValueOnce
  fromValue = fromValueOnce

instance {-# OVERLAPPABLE #-} Valuable a => Valuable [a] where
  toValue = toValueOnce . map toValue
  fromValue = map fromValue . fromValueOnce

instance {-# OVERLAPPABLE #-} (Valuable a, Valuable b) => Valuable (a, b) where
  toValue (a, b) = toValueOnce (toValue a, toValue b)
  fromValue v = let (va, vb) = fromValueOnce v
                in (fromValue va, fromValue vb)

instance {-# OVERLAPPABLE #-} (Typeable a, Show a) => Valuable a where
  toValue = toValueOnce
  fromValue = fromValueOnce
